import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:state_tree_interface/state_tree_interface.dart' as i;

export 'package:state_tree_interface/state_tree_interface.dart'
    show Failure, Late, Async, LateAsync, R, Success, Result, Pending, Loading;

class _StateTreeLeaf extends i.StateTreeLeaf {
  _StateTreeLeaf(super.parent);

  @override
  StateTreeNode get parent => super.parent as StateTreeNode;
}

class _ProviderLeaf<T> extends _StateTreeLeaf {
  T value;
  i.Context<T> context;

  _ProviderLeaf(super.parent, this.value, this.context);

  @override
  void init() {
    super.init();

    if (parent._element._providerLeafs.containsKey(context)) {
      throw StateError('$context already exists');
    }

    parent._element._providerLeafs[context] = this;
  }

  @override
  void dispose() {
    parent._element._providerLeafs.remove(context);
    super.dispose();
  }
}

class _ConsumerLeaf<T, K> extends _StateTreeLeaf {
  final i.Context<K> context;
  final T Function(K value)? _query;
  late T value;

  final bool subscribe;

  _ConsumerLeaf(
    super.parent,
    this._query, {
    required this.context,
    this.subscribe = true,
  });

  @override
  void init() {
    super.init();
    value = query(
      _Provider.of<T, K>(
        parent._element,
        aspect: this,
        context: context,
        subscribe: subscribe,
      ).value,
    );
  }

  T _identity(K value) => value as T;

  T query(K value) {
    return (_query ?? _identity)(value);
  }

  bool shouldUpdate(K oldValue, K newValue) {
    final a = query(oldValue);
    final b = query(newValue);

    final shouldUpdate = a != b;

    if (shouldUpdate) {
      value = b;
    }

    return shouldUpdate;
  }
}

class _CallLeaf<T, K> extends _StateTreeLeaf {
  final T Function(K arg) fn;
  final K arg;
  final InvocationMode mode;
  late T value;

  _CallLeaf(super.parent, this.fn, this.arg, this.mode);

  @override
  void init() {
    super.init();

    switch (mode) {
      case InvocationMode.sync:
        value = fn(arg);
        break;
      case InvocationMode.postFrame:
        value = null as T;
        WidgetsBinding.instance.addPostFrameCallback((_) {
          fn(arg);
        });
        break;
    }
  }
}

class _DisposeLeaf<T> extends _StateTreeLeaf {
  final void Function(T value) onDispose;
  final T dep;

  _DisposeLeaf(super.parent, this.onDispose, this.dep);

  @override
  void dispose() {
    onDispose(dep);
    super.dispose();
  }
}

class _AsyncLeaf<T> extends _StateTreeLeaf {
  final Future<T> future;
  late i.Async<T> value;

  _AsyncLeaf(super.parent, this.future);

  bool _shouldNotify = true;

  @override
  void init() {
    super.init();

    value = i.Async.loading();

    future.then((value) {
      if (this.value is i.Pending) {
        this.value = i.Async.success(value);

        if (!_shouldNotify) return;
        if (!parent._element.mounted) return;
        parent._element.markNeedsBuild();
      }
    }).catchError((error, stackTrace) {
      if (this.value is i.Pending) {
        this.value = i.Async.failure(error, stackTrace);

        if (!_shouldNotify) return;
        if (!parent._element.mounted) return;
        parent._element.markNeedsBuild();
      }
    });
  }

  @override
  void dispose() {
    _shouldNotify = false;
    super.dispose();
  }
}

class _StreamLeaf<T> extends _StateTreeLeaf {
  final Stream<T> stream;
  late i.LateAsync<T> value;
  late StreamSubscription subscription;

  _StreamLeaf(super.parent, this.stream);

  @override
  void init() {
    super.init();

    value = i.LateAsync.pending();

    subscription = stream.listen((value) {
      if (this.value is i.Pending) {
        this.value = i.LateAsync.success(value);

        if (!parent._element.mounted) return;
        parent._element.markNeedsBuild();
      }
    }, onError: (error, stackTrace) {
      if (this.value is i.Pending) {
        this.value = i.LateAsync.failure(error, stackTrace);

        if (!parent._element.mounted) return;
        parent._element.markNeedsBuild();
      }
    });
  }

  @override
  void dispose() {
    subscription.cancel();
    super.dispose();
  }
}

class InvocationMode implements i.InvocationMode {
  const InvocationMode._();

  static const sync = _Sync();
  static const postFrame = _PostFrame();
}

class _Sync extends InvocationMode {
  const _Sync() : super._();
}

class _PostFrame extends InvocationMode {
  const _PostFrame() : super._();
}

class StateTreeNode<U> implements i.StateTreeNode<U> {
  final _StateTreeWidgetElement _element;

  StateTreeNode(this._element);

  late U _value;
  bool _isInitialized = false;

  U receive() {
    return _value;
  }

  U send(U value) {
    if (_element._isBuilding) {
      if (!_isInitialized) {
        _value = value;
        _isInitialized = true;
      }

      return _value;
    }

    if (!_element.mounted) return _value;

    _value = value;
    _element.markNeedsBuild();
    return value;
  }

  @override
  T init<T>(T Function() factory) {
    return call<T, Object>((_) {
      return factory();
    }, const Object());
  }

  @override
  T call<T, K>(
    T Function(K arg) fn,
    K dep, {
    i.InvocationMode mode = InvocationMode.sync,
  }) {
    final leaf = _CallLeaf<T, K>(this, fn, dep, mode as InvocationMode);
    return _element._next(leaf, dep ?? const Object()).value;
  }

  @override
  void dispose<T>(void Function(T value) onDispose, T dep) {
    final leaf = _DisposeLeaf<T>(this, onDispose, dep);
    _element._next(leaf, dep as Object);
  }

  i.Async<K> future<K>(Future<K> future) {
    final leaf = _AsyncLeaf<K>(this, future);
    return _element._next(leaf, future).value;
  }

  i.LateAsync<K> stream<K>(Stream<K> stream) {
    final leaf = _StreamLeaf<K>(this, stream);
    return _element._next(leaf, stream).value;
  }

  @override
  K watch<K>(i.Context<K> context) {
    final leaf = _ConsumerLeaf(this, null, context: context);
    return _element._next(leaf, const Object()).value;
  }

  @override
  K query<K, T>(i.Context<T> context, K Function(T value) query) {
    final leaf = _ConsumerLeaf<K, T>(this, query, context: context);
    return _element._next(leaf, const Object()).value;
  }

  @override
  K read<K>(i.Context<K> context) {
    return _Provider.of<U, K>(_element, context: context).value;
  }
}

abstract class StateTreeWidget<T> extends Widget {
  const StateTreeWidget({super.key});

  Widget build(BuildContext context, StateTreeNode<T> n);

  @override
  Element createElement() {
    return _StateTreeWidgetElement<T>(this);
  }
}

class _StateTreeWidgetElement<T> extends ComponentElement {
  late final node = StateTreeNode<T>(this);
  int cursor = 0;
  final _leafs = <(Object, i.StateTreeLeaf)>[];
  final _providerLeafs = <i.Context, _ProviderLeaf>{};

  _StateTreeWidgetElement(StateTreeWidget widget) : super(widget);

  @override
  StateTreeWidget get widget => super.widget as StateTreeWidget;

  bool _isBuilding = false;

  T _next<T extends i.StateTreeLeaf>(T leaf, Object newKey) {
    if (!_isBuilding) {
      throw StateError(
        '$leaf could only be used synchronously during build.',
      );
    }

    if (cursor < _leafs.length) {
      final (dep, existing) = _leafs[cursor];
      if (dep == newKey) {
        cursor++;
        return existing as T;
      } else {
        final newLeaf = existing.update(leaf, newKey);
        _leafs[cursor] = (newKey, newLeaf);
        cursor++;
        return newLeaf as T;
      }
    }

    leaf.init();
    _leafs.add((newKey, leaf));
    cursor++;

    return leaf;
  }

  @override
  Widget build() {
    cursor = 0;

    _isBuilding = true;
    var child = widget.build(this, node);
    _isBuilding = false;

    for (int j = cursor; j < _leafs.length; j++) {
      final (_, leaf) = _leafs[j];
      leaf.dispose();
    }

    _leafs.removeRange(cursor, _leafs.length);

    if (_providerLeafs.isNotEmpty) {
      child = _Provider(
        leafs: {..._providerLeafs},
        parentElement: this,
        child: child,
      );
    }

    return child;
  }

  @override
  void unmount() {
    for (final (_, leaf) in _leafs) {
      leaf.dispose();
    }

    super.unmount();
  }
}

class _Provider extends InheritedModel<_ConsumerLeaf> {
  final Map<i.Context, _ProviderLeaf> leafs;
  final Element parentElement;

  const _Provider({
    required super.child,
    required this.leafs,
    required this.parentElement,
  });

  @override
  bool updateShouldNotify(_Provider oldWidget) {
    final oldKeys = oldWidget.leafs.keys.toSet();
    final newKeys = leafs.keys.toSet();

    final shouldNotify = oldKeys.difference(newKeys).isNotEmpty;

    if (shouldNotify) return shouldNotify;

    for (final key in leafs.keys) {
      final oldLeaf = oldWidget.leafs[key];
      final newLeaf = leafs[key];

      if (oldLeaf!.value != newLeaf!.value) {
        return true;
      }
    }

    return false;
  }

  @override
  bool updateShouldNotifyDependent(
    _Provider oldWidget,
    Set<_ConsumerLeaf> dependencies,
  ) {
    final shouldNotify = dependencies.any((e) {
      final key = e.context;

      final oldLeaf = oldWidget.leafs[key];
      final newLeaf = leafs[key];

      if (oldLeaf == null || newLeaf == null) return true;

      return e.shouldUpdate(oldLeaf.value, newLeaf.value);
    });

    return shouldNotify;
  }

  static _ProviderLeaf<K> of<T, K>(
    Element element, {
    _ConsumerLeaf<T, K>? aspect,
    bool subscribe = false,
    Object? context,
  }) {
    Element? e = element;

    while (true) {
      final ie = e?.getElementForInheritedWidgetOfExactType<_Provider>();
      final model = ie?.widget as _Provider?;

      if (model == null) {
        throw StateError('No $context found');
      }

      final leaf = model.leafs[context];

      if (leaf == null) {
        e = model.parentElement;
        continue;
      }

      if (subscribe) element.dependOnInheritedElement(ie!, aspect: aspect);

      return leaf as _ProviderLeaf<K>;
    }
  }
}

class StateTreeBuilder<T> extends StateTreeWidget<T> {
  final Widget Function(
    BuildContext context,
    StateTreeNode<T> n,
    Widget child,
  ) builder;

  final Widget child;

  const StateTreeBuilder(
    this.builder, {
    Key? key,
    this.child = const SizedBox(),
  }) : super(key: key);

  @override
  Widget build(BuildContext context, StateTreeNode<T> n) {
    return builder(context, n, child);
  }
}

class Context<T> extends i.Context<T> {
  const Context(super.token);

  @override
  void provide(i.StateTreeNode n, T value, [Object? key]) {
    final leaf = _ProviderLeaf<T>(n, value, this);
    (n as StateTreeNode)._element._next(leaf, key ?? (value,));
  }
}

class Impl<T> {
  final StateTreeNode<T> n;
  const Impl(this.n);

  T get value => n.receive();
  set value(T value) => n.send(value);
}
